/*  Copyright (C) 2000, 2006  Marc Toussaint (mtoussai@inf.ed.ac.uk)
    under the terms of the GNU LGPL (http://www.gnu.org/copyleft/lesser.html)
    see the `std.h' file for a full copyright statement  */

#include"dynamics.h"

double MT::phaseDist(double x0,double v0,double x1,double v1){
  double x,v,dx,dv,xc;
  x=.5*(x0+x1);
  v=.5*(v0+v1);
  dx=x0-x1;
  dv=v0-v1;
  if(fabs(dx)<1e-6) if(MT::sign(v0)!=MT::sign(v1)) return MT_PI; else return MT_2PI;
  xc=x+v*dv/dx;

  double phi1=MT::phi(x1-xc,v1);
  double phi2=MT::phi(x0-xc,v0);
  double phi=phi2-phi1;
  if(phi<0.) phi+=MT_2PI;
  if(phi>MT_2PI) phi-=MT_2PI;
  return phi;
}


double MT::phaseAccel(double x0,double v0,double x1,double v1,double maxA){
  double x,v,dx,dv,xc,a;
  x=.5*(x0+x1);
  v=.5*(v0+v1);
  dx=x0-x1;
  dv=v0-v1;
  if(fabs(dx)<1e-6){
    if(MT::sign(v0)!=MT::sign(v1)) if(v0>0.) return -maxA; else return maxA; //return 0;
    else if(v0<0.) return -maxA; else return maxA;
  }
  xc=x+v*dv/dx;
  std::cout <<"null x = "<<xc <<std::endl;
  a=-(x0-xc);

  if(fabs(v0)>fabs(v1)){
  if(dx>0 && dv>0. && a>0.) return .0;//
  if(dx<0 && dv<0. && a<0.) return .0;//
  }
  a*=1.2;
  if(maxA){
    if(a> maxA) return  maxA;
    if(a<-maxA) return -maxA;
  }
  return a;
}

void MT::phaseTangent(double x0,double v0,double x1,double v1,double& dx,double& dv,double maxA){
  dx=v0;
  dv=phaseAccel(x0,v0,x1,v1,maxA);
}

