#include "functions.h"

double NNsdv(const double& a,const double& b,double sdv){
  double d=(a-b)/sdv;
  double norm = 1./(::sqrt(MT_2PI)*sdv);
  return norm*EXP(-.5*d*d);
}
double NNsdv(double x,double sdv){
  x/=sdv;
  double norm = 1./(::sqrt(MT_2PI)*sdv);
  return norm*EXP(-.5*x*x);
}
double NNinv(const doubleA& a,const doubleA& b,const doubleA& invC){
  doubleA d=a-b;
  double norm = ::sqrt(determinant((1./MT_2PI)*invC));
  return norm*EXP(-.5*scalarProduct(invC,d,d));
}
double logNNinv(const doubleA& a,const doubleA& b,const doubleA& invC){
  doubleA d=a-b;
  double norm = ::sqrt(fabs(determinant((1./MT_2PI)*invC)));
  return ::log(norm) + (-.5*scalarProduct(invC,d,d));
}
double logNN(const doubleA& a,const doubleA& b,const doubleA& C){
  return logNNinv(a,b,inverse(C));
}
double NN(const doubleA& a,const doubleA& b,const doubleA& C){
  return NNinv(a,b,inverse(C));
}
//! non-normalized!! Gaussian function (f(0)=1)
double NNNNinv(const doubleA& a,const doubleA& b,const doubleA& invC){
  doubleA d=a-b;
  return EXP(-.5*scalarProduct(invC,d,d));
}
double NNNN(const doubleA& a,const doubleA& b,const doubleA& C){
  return NNNNinv(a,b,inverse(C));
}
double NNzeroinv(const doubleA& x,const doubleA& invC){
  double norm = ::sqrt(determinant((1./MT_2PI)*invC));
  return norm*EXP(-.5*scalarProduct(invC,x,x));
}
//! gradient of a Gaussian
double dNNinv(const doubleA& x,const doubleA& a,const doubleA& invA,doubleA& grad){
  double y=NNinv(x,a,invA);
  grad = y * invA * (a-x);
  return y;
}
//! gradient of a non-normalized Gaussian
double dNNNNinv(const doubleA& x,const doubleA& a,const doubleA& invA,doubleA& grad){
  double y=NNNNinv(x,a,invA);
  grad = y * invA * (a-x);
  return y;
}
double NNsdv(const doubleA& a,const doubleA& b,double sdv){
  double norm = 1./(::sqrt(MT_2PI)*sdv);
  return norm*EXP(-.5*sqrDistance(a,b)/(sdv*sdv));
}
double NNzerosdv(const doubleA& x,double sdv){
  double norm = 1./(::sqrt(MT_2PI)*sdv);
  return norm*EXP(-.5*sumOfSqr(x)/(sdv*sdv));
}
