#include "gauss.h"

#define takeC 1
#define takeU 2

int decideCU(const Gaussian &a,const Gaussian&b){
  if(a.okU && b.okU){
    return takeU;
  }
  if(a.okC && b.okC){
    return takeC;
  }
  if(a.okC && !a.okU && !b.okC && b.okU){
    double Za=minDiag(a.C),Zb=minDiag(b.U);
    if(Za>Zb){ a.makeU(); return takeU; }else{ b.makeC(); return takeC; }
  }
  if(!a.okC && a.okU && b.okC && !b.okU){
    double Za=minDiag(a.U),Zb=minDiag(b.C);
    if(Za>Zb){ a.makeC(); return takeC; }else{ b.makeU(); return takeU; }
  }
  NIY;
}

void Gaussian::operator=(const Gaussian& x){
  if(x.okC){ c=x.c; C=x.C; }
  if(x.okU){ u=x.u; U=x.U; }
  okC=x.okC; okU=x.okU;
}

void Gaussian::makeU() const{ //the const is actually not true -- but oh well..
  if(okU) return;
  CHECK(okC,"can't make canonical rep without normal rep");
  //CHECK(minDiag(C)>1e-10,"inverting zero variance matrix");
  for(uint i=0;i<C.d0;i++){ if(C(i,i)<1e-10) C(i,i)=1e-10; }
  Gaussian *g=(Gaussian*)this; //get a non-const pointer (hack :-()
  inverse(g->U,C);
  g->u=U*c;
  g->okU=true;
}

void Gaussian::makeC() const{ //the const is actually not true -- but oh well..
  if(okC) return;
  CHECK(okU,"can't make canonical rep without normal rep");
  //CHECK(minDiag(U)>1e-10,"inverting zero precision matrix");
  for(uint i=0;i<U.d0;i++){ if(U(i,i)<1e-10) U(i,i)=1e-10; }
  Gaussian *g=(Gaussian*)this; //get a non-const pointer (hack :-()
  inverse(g->C,U);
  g->c=C*u;
  g->okC=true;
}

void Gaussian::setConditional(uint n,double var1,double var2){
  arr d,dd; d.setDiag(var1,n); dd.setDiag(var1+var2,n);
  blockMatrix(C,d,d,d,dd);
  c.resize(2*n);
  c.setZero();
  setCU(true,false);
}

void Gaussian::setDiagonal(uint n,double var){
  c.resize(n); c.setZero();
  C.setDiag(var,n);
  setCU(true,false);
}

void Gaussian::setConditional(const arr& f,const arr& F,const arr& Q){
  uint n1=F.d0;
  //uint n2=F.d1;
  //uint n=n1+n2;
  Gaussian a;
  //a.setUniform(n1);
  a.setDiagonal(n1,100.);
  a.makeC();
  arr FA=F*a.C,tFA=~FA;
  blockMatrix(C, a.C, tFA, FA, Q + F*tFA);
  blockVector(c, a.c, F*a.c + f);
  setCU(true,false);
}

void Gaussian::write(std::ostream& os) const{
  if(okC){
    os <<c <<C;
  }else{
    CHECK(okU,"");
    os <<u <<U;
  }
}
void Gaussian::read(std::istream& is){
  is >>c >>C;
  setCU(true,false);
}

void estimate(Gaussian& g,const arr& X){
  uint k=X.d0;
  double Z=(double)k;
  arr ones(k); ones=1.;
  g.c=ones*X;
  g.C =~X*X;
  g.C =(g.C  - (g.c^g.c)/Z)/Z;
  g.c/=Z;
  g.setCU(true,false);
}

void estimateWeighted(Gaussian& g,const arr& X,const arr& W){
  CHECK(X.d0==W.N,"");
  uint k=X.d0,n=X.d1,i;
  g.c.resize(n);   g.c.setZero();
  g.C.resize(n,n); g.C.setZero();
  double Z=0.;
  arr X_i;
  for(i=0;i<k;i++){
    Z   += W(i);
    X_i.referToSubDim(X,i);
    g.c += W(i) * X_i;
    g.C += W(i) * (X_i ^ X_i);
  }
  g.C =(g.C  - (g.c^g.c)/Z)/Z;
  g.c/=Z;
  g.setCU(true,false);
}

void collapseMoG(Gaussian& g,const arr& P,const gaussA& G){
  G(0).makeC();
  uint i,k=P.N,n=G(0).c.N;
  g.c.resize(n);   g.c.setZero();
  g.C.resize(n,n); g.C.setZero();
  arr CC(n,n);      CC.setZero();
  double Z=0.;
  for(i=0;i<k;i++){
    G(i).makeC();
    Z   += P(i);
    g.c += P(i) * G(i).c;
    g.C += P(i) * G(i).C;
    CC  += P(i) * (G(i).c ^ G(i).c);
  }
  g.C =(g.C + CC - (g.c^g.c)/Z)/Z;
  g.c/=Z;
  g.setCU(true,false);
}

void collapseMoG(Gaussian& g,const arr& P,const MT::Array<Gaussian*>& G){
  G(0)->makeC();
  uint i,k=P.N,n=G(0)->c.N;
  g.c.resize(n);   g.c.setZero();
  g.C.resize(n,n); g.C.setZero();
  arr CC(n,n);      CC.setZero();
  double Z=0.;
  for(i=0;i<k;i++){
    G(i)->makeC();
    Z   += P(i);
    g.c += P(i) * G(i)->c;
    g.C += P(i) * G(i)->C;
    CC  += P(i) * (G(i)->c ^ G(i)->c);
  }
  g.C =(g.C + CC - (g.c^g.c)/Z)/Z;
  g.c/=Z;
  g.setCU(true,false);
}

void resampleAndEstimate(Gaussian& g,double (*f)(const arr& x),uint N){
  doubleA X,W(N);
  double Z;
  uint i;
  sample(X,N,g);
  Z=0.;
  for(i=0;i<X.d0;i++){
    W(i) = f(X[i]);
    Z += W(i);
  }
  CHECK(Z>1e-10,"sample weights sum to approx. zero");
  estimateWeighted(g,X,W);
};

void sample(arr& x,const Gaussian& g){
  g.makeC();
  x.resize(g.C.d0);
  rndGauss(x);
  arr U,V;
  svd(g.C,U,V);
  x = U*x;
  x += g.c;
}

void sample(arr& X,uint N,const Gaussian& g){
  uint i;
  X.resize(N,g.c.N);
  for(i=0;i<N;i++) sample(X[i](),g);
}

void systematicWeightedSamples(arr& X,arr& W,const Gaussian& g){
  uint n=g.c.N,i;

  //parameters
  double alpha=1., kappa=2.;
  double lambda=alpha*alpha*(n+kappa)-n;

  //standard samples
  g.makeC();
  X.resize(2*n+1,n);
  W.resize(X.d0);
  arr U,V,d(n),X_i;
  svd(g.C,U,V);
  X_i.referToSubDim(X,0);        X_i=g.c;    W(0)=(lambda/(n+lambda));
  for(i=0;i<n;i++){
    d.setZero();
    d(i)=sqrt(n+lambda);
    d=U*d;
    //d=U.sub(0,-1,i,i); //i-th column
    //d.resize(n);
    X_i.referToSubDim(X,2*i+1);  X_i=g.c+d;  W(2*i+1)=1./(2.*(n+lambda));
    X_i.referToSubDim(X,2*i+2);  X_i=g.c-d;  W(2*i+2)=1./(2.*(n+lambda));
  }
}

void product(Gaussian& x, const Gaussian& a, const Gaussian& b,double *logNorm){
  int takeCU=decideCU(a,b);
  //a.makeC(); b.makeC(); takeCU=takeC;
  a.makeU(); b.makeU(); takeCU=takeU;
  if(takeCU==takeC){
    arr X,AX,BX;
    inverse(X,a.C + b.C);
    if(logNorm){
      //cout <<"computing norm:\na=" <<a.c <<"b=" <<b.c <<"\nX=" <<inverse(X) <<endl;
      *logNorm = logNNinv(a.c,b.c,X);
    }
    AX = a.C*X;
    BX = b.C*X;
    x.C = AX*b.C;          //=A*X*B = A * (A+B)^-1 * B;
    x.c = BX*a.c + AX*b.c; //=B*X*a + A*X*b;
    x.setCU(true,false);
  }else{
    x.U = a.U + b.U;
    if(logNorm){
      a.makeC();
      b.makeC();
      //cout <<"computing norm:\na=" <<a.c <<"b=" <<b.c <<"\nX=" <<inverse(X) <<endl;
      *logNorm = logNNinv(a.c,b.c,inverse(a.C + b.C));
    }
    x.u = a.u + b.u;
    x.setCU(false,true);
  }
}

void division(Gaussian& x, const Gaussian& a, const Gaussian& b,double *logNorm){
  int takeCU=decideCU(a,b);
  //a.makeC(); b.makeC(); takeCU=takeC;
  if(takeCU==takeC){
    arr X,AX,BX;
    X = inverse(a.C - b.C);  //X=(A-B)^-1
    AX = a.C*X;
    BX = b.C*X;
    x.C = AX*(-b.C);         //C=A*X*(-B);
    x.c = AX*b.c - BX*a.c;   //c=A*X*b - B*X*a;
    if(logNorm){
      a.makeC();
      b.makeC();
      *logNorm = - logNN(x.c,b.c,x.C+b.C);
    }
    x.setCU(true,false);
  }else{
    x.U = a.U - b.U;
    x.u = a.u - b.u;
    if(logNorm){
      a.makeC();
      b.makeC();
      x.makeC();
      *logNorm = - logNN(x.c,b.c,x.C+b.C);
    }
    x.setCU(false,true);
  }
}

void forward(Gaussian& y, const Gaussian& x, arr& f, arr& F, arr& Q){
  x.makeC();
  y.c = F*x.c + f;
  y.C = F*x.C*~F + Q;
  y.setCU(true,false);
}

void backward(Gaussian& x, const Gaussian& y, arr& f, arr& F, arr& Q,double updateStep){
  CHECK(y.c.N==f.N && y.c.N==F.d0 && y.c.N==Q.d0,"linBwd: output dimension does not match lin fwd fct"); 
  y.makeC();
  arr invF=inverse(F);
  if(F.d0>=F.d1){ //non-redundant map
    x.c = invF*(y.c-f);
    x.C = invF*(y.C+Q)*~invF;
  }else{
    CHECK(x.c.N==F.d1,"x.c must be initialized when pulling back a redundant mapping!");
    //arr tF = ~F;
    //arr iF = tF * inverse(F*tF);
    //CHECK(maxDiff(invF,iF)<1e-10,"pseudo inverse not right");
    arr D; D.setId(F.d1);
    x.c += updateStep*invF*(y.c - (f + F*x.c));
    x.C = invF*(y.C+Q)*~invF;

    //strech the redundant dimensions
    arr U,w,V,Vt,vk;
    svd(F,U,w,V);
    double scale=norm(w);
    transpose(Vt,V);
    //cout <<U <<w <<V;
    for(uint k=F.d0;k<w.N;k++){ //eigen vectors with zero eigen value
      CHECK(w(k)<1e-10*scale,"eigen value should be zero!");
      vk.referToSubDim(Vt,k);
      x.C += (1e+0*scale)* (vk^vk);
    }
  }
  x.setCU(true,false);
}

/*void forward(Gaussian& y, const Gaussian& x, Trans f, arr& Q){
  x.makeC();
  y=x;
  unscentedTransform(y.c,y.C,f);
  if(y.C.d0==2*Q.d0 && y.C.d1==2*Q.d1){ //assume joint is propagated
    arr zero(Q.d0,Q.d1); zero.setZero();
    arr block; blockMatrix(block,Q,zero,zero,Q);
    y.C += block;
  }else{
    y.C += Q;
  }
  y.okC=true; y.okU=false;
}*/

/*void fctFwd(Gaussian& y,Gaussian& x, Trans f){
  y=x;
  unscentedTransform(y.c,y.C,f);
}*/

void unscentedTransform(Gaussian &b, const Gaussian &a, Trans f){
  uint i,n=a.c.N;

  //following Murphy's http://www.cs.ubc.ca/~murphyk/Papers/dbnchapter.pdf

  //parameters
  double alpha=1., beta=0., kappa=2.;
  double lambda=alpha*alpha*(n+kappa)-n;

  //standard samples
  b.makeC();
  MT::Array<arr> X(2*n+1);
  arr U,V,d(n);
  svd(a.C,U,V);
  X(2*n)=a.c;
  for(i=0;i<n;i++){
    d=U.sub(0,-1,i,i); //i-th column
    d.resize(n);
    d*=sqrt(n+lambda);
    X(2*i)=a.c+d;
    X(2*i+1)=a.c-d;
  }

  //transform
  for(i=0;i<X.d0;i++) f(X(i));

  //weights
  double Wm0,Wc0,Wmc;
  Wm0=(lambda/(n+lambda));
  Wc0=(lambda/(n+lambda)) + (1-alpha*alpha+beta);
  Wmc=1./(2.*(n+lambda));

  //new means and covariance
  b.c.resize(n); b.C.resize(n,n);
  b.c = Wm0 * X(2*n);
  for(i=0;i<2*n;i++) b.c += Wmc * X(i);
  b.C = Wc0 * (X(2*n)-b.c)^(X(2*n)-b.c);
  for(i=0;i<2*n;i++) b.C += Wmc * (X(i)-b.c)^(X(i)-b.c);

#if 0
  std::cout <<"unscent transform : " 
    <<a <<b <<A <<B
    <<"error : " <<sqrDistance(a,b) <<' ' <<sqrDistance(A,B) <<std::endl;
  
  plotCovariance(a,A);
  plotPoints(X);
  plotFlush();

  //re-estimate
  arr b,B;
  estimate(X,b,B);
  //B*=double(2*n+1)/(2*n);

  //re-estimate
  estimate(X,a,A);
  A*=double(2*n+1)/(2*n); //to account for the mean as a sample point
#endif
}

void getLinFwdFromJoint(arr& f, arr& F, arr& Q, uint n1, uint n2, Gaussian& x){
  CHECK(n1+n2==x.c.N,"");
  arr A,B,C,invA,a,b;
  x.makeC();
  a=x.c.sub(0,n1-1);        //upper vector
  b=x.c.sub(n1,-1);         //lower vector
  A=x.C.sub(0,n1-1,0,n1-1); //upper left
  B=x.C.sub(n1,-1,n1,-1);   //lower right
  C=x.C.sub(0,n1-1,n1,-1);  //upper right
  invA=inverse(A);

  F = ~C*invA;
  f = b - F*a;
  Q = B - F*C;
}

void getMarginal(Gaussian& x,const Gaussian& joint,uint lo,uint hi){
  joint.makeC();
  CHECK(lo<=hi && hi<=joint.c.N,"");
  x.c=joint.c.sub(lo,hi);       //sub vection
  x.C=joint.C.sub(lo,hi,lo,hi); //sub block matrix
  x.setCU(true,false);
}

void getMarginal(Gaussian& x,const Gaussian& joint,uintA& list){
  uint n=list.N,i,j;
  if(joint.okC){
    x.c.resize(n);
    x.C.resize(n,n);
    for(i=0;i<n;i++){
      x.c(i)=joint.c(list(i));
      for(j=0;j<n;j++) x.C(i,j)=joint.C(list(i),list(j));
    }
    x.setCU(true,false);
  }else{
    //--- first just extract the components u=(a,b) and U=(A,C;~C,B)
    
    //set permutation pi complementary to list
    uintA pi;
    if(joint.N()>list.N){
      pi.setStraightPerm(joint.N());
      for(i=0;i<list.N;i++) pi.permute(i,list(i));
      pi = pi.sub(n,-1);
    }else{
      pi.resize(0);
    }
    uint m=pi.N;

    //extract components
    arr A(n,n),B(m,m),C(n,m),a(n),b(m);
    for(i=0;i<n;i++){
      a(i)=joint.u(list(i));
      for(j=0;j<n;j++) A(i,j)=joint.U(list(i),list(j));
    }
    for(i=0;i<m;i++){
      b(i)=joint.u(pi(i));
      for(j=0;j<m;j++) B(i,j)=joint.U(pi(i),pi(j));
    }
    for(i=0;i<n;i++) for(j=0;j<m;j++) C(i,j)=joint.U(list(i),pi(j));

    //--- now it's easy:
    arr CinvB=C*inverse(B);
    x.u = a - CinvB * b;
    x.U = A - CinvB * ~C;
    x.setCU(false,true);
  }
}

void setMarginal(Gaussian& joint,const Gaussian& marg,uintA& list){
  uint n=list.N,i,j;
  marg.makeC();
  joint.makeC();
  if(marg.okC){
    CHECK(marg.c.N==n && marg.C.d0==n,"");
    for(i=0;i<n;i++){
      joint.c(list(i))=marg.c(i);
      for(j=0;j<n;j++) joint.C(list(i),list(j))=marg.C(i,j);
    }
    joint.setCU(true,false);
  }else{
    marg.makeC();
    CHECK(marg.c.N==n && marg.C.d0==n,"");
    for(i=0;i<n;i++){
      joint.c(list(i))=marg.c(i);
      for(j=0;j<n;j++) joint.C(list(i),list(j))=marg.C(i,j);
    }
    joint.setCU(true,false);
  }
}

void getMarginalsFromJoint(Gaussian& x,Gaussian& y,Gaussian& xi){
  xi.makeC();
  uint n=xi.c.N/2;
  CHECK(xi.c.N=n+n,"");
  x.c=xi.c.sub(0,n-1);        //upper vector
  y.c=xi.c.sub(n,-1);         //lower vector
  x.C=xi.C.sub(0,n-1,0,n-1); //upper left
  y.C=xi.C.sub(n,-1,n,-1);   //lower right
  x.setCU(true,false);
  y.setCU(true,false);
}

void joinMarginalAndConditional(Gaussian& xi,Gaussian& a, arr& f, arr& F, arr& Q){
  a.makeC();
  arr FA=F*a.C,tFA=~FA;
  blockMatrix(xi.C, a.C, tFA, FA, Q + F*tFA);
  blockVector(xi.c, a.c, F*a.c + f);
  xi.setCU(true,false);
}

//! xi is P(x,y), returns P(y|x)={Fx+f,Q}, cut is the lower index of y
void getConditional(const Gaussian& xi,uint cut,arr& f,arr& F,arr& Q){
  arr A,B,C,a,b;
  xi.makeC();
  uint k=cut,l=xi.C.d0-k;
  A.resize(k,k);
  B.resize(l,l);
  C.resize(k,l);
  xi.C.getBlock(A,0,0);
  xi.C.getBlock(B,k,k);
  xi.C.getBlock(C,0,k);

  a.resize(k);
  b.resize(l);
  xi.c.getBlock(a,0);
  xi.c.getBlock(b,k);

  F = ~C * inverse(A);
  f = b - F*a;
  Q = B - F*C;
}

//simple version with first product rule
void multiplyToJoint(Gaussian& xi, Gaussian& b){
  uint n=b.c.N;
  CHECK(xi.c.N==2*n,"only for doubles sizes yet");
  arr zero(n,n); zero.setZero();
  arr big; big.setDiag(1e8,n);
  Gaussian B;
  B.c.setUni(0.,n); B.c.append(b.c);
  blockMatrix(B.C,big,zero,zero,b.C);
  product(xi,xi,B);
}

/* version with 2nd product rule
void multiplyToJoint(Gaussian& xi, Gaussian& b){
  uint i,j,n2=b.c.N,n1=xi.c.N-b.c.N;
  arr invA,invB,invC,ac=xi.c;
  invA=inverse(xi.C);
  invB=inverse(b.C);
  //construct sum of both
  invC=invA;
  for(i=0;i<n2;i++) for(j=0;j<n2;j++){ //loop over the lower right blockmatrix
    invC(n1+i,n1+j) += invB(i,j);
  }
  xi.C = inverse(invC);
  xi.c.resize(n1);
  xi.c.setZero();
  xi.c.append(invB*b.c); //now is a block vector xi.c = [0 invB*b]
  xi.c = xi.C * (invA * ac + xi.c);
}
*/

bool sameGaussian(const Gaussian &a,const Gaussian &b){
#if 0
  a.makeC(); b.makeC(); a.makeU(); b.makeU();
  return (maxDiff(a.C,b.C)<1e-3 && maxDiff(a.c,b.c)<1e-3) || (maxDiff(a.U,b.U)<1e-3 && maxDiff(a.u,b.u)<1e-3);
#elif 0
  int takeCU=decideCU(a,b);
  if(takeCU==takeC){
    return (maxDiff(a.C,b.C)<1e-8) && maxDiff(a.c,b.c)<1e-8;
  }else{
    return (maxDiff(a.U,b.U)<1e-8) && maxDiff(a.u,b.u)<1e-8;
  }
#else
  double d=KLDsym(a,b);
  //DEBUGINF(cout <<"KLD = " <<d <<endl;);
  return d<1.;
#endif
}

double KLDsym(const Gaussian &a,const Gaussian &b){
  double x=0.;
  a.makeU(); b.makeU();
  a.makeC(); b.makeC();
  x = trace(b.U*a.C) + trace(a.U*b.C) + sqrDistance(a.U+b.U,a.c,b.c) - (2.*a.c.N);
  x /= 4.;
  return x;
}

void write(gaussA& x,const char* name){
  ofstream os;
  MT::open(os,name);
  MT::IOraw=true;
  for(uint i=0;i<x.N;i++){
    x(i).makeC();
    os <<x(i).c <<" \t ";
    x(i).C.write(os," "," \t ");
    os <<endl;
  }
}



#if 0
//===========================================================================
//
// older but correct versions
//


void equal(arr& c, arr& C, arr& a, arr& A){
  c=a;
  C=A;
}

void unscentedTransform(arr& a,arr& A,Trans f);

//! multiplication: {a,A} * {b,B} = {c,C} * norm
void product(arr& c, arr& C, arr& a, arr& A, arr& b, arr& B,double *norm=0){
  CHECK(&C != &B,"only C and A can be the same matrix");
  arr X,Y;
  X = inverse(A + B);
  Y = A*X;
  C = Y*B;   //=A*X*B;
  c = B*X*a + Y*b; //=B*X*a + A*X*b;
  if(norm){
    HALT("compute also normalization contant depending on a and b");
  }
}

//! given x~{a,A} and y|x~{Fx+f,Q}, this returns y~{c,C}
void linFwd(arr& c, arr& C, arr& a, arr& A, arr& f, arr& F, arr& Q){
  c = F*a + f;
  C = F*A*~F + Q;
}

//! given x~{a,A} and y|x~{f(x),Q}, this returns y~{c,C}
void fctFwd(arr& c, arr& C, arr& a, arr& A, Trans f, arr& Q){
  c=a; C=A;
  unscentedTransform(c,C,f);
  if(C.d0==2*Q.d0 && C.d1==2*Q.d1){ //assume joint is propagated
    arr zero(Q.d0,Q.d1); zero.setZero();
    arr block; blockMatrix(block,Q,zero,zero,Q);
    C += block;
  }else{
    C += Q;
  }
}

//! given y~{b,B} and y|x~{Fx+f,Q}, this returns x~{c,C}
void linBwd(arr& c, arr& C, arr& b, arr& B, arr& f, arr& F, arr& Q){
  arr invF=inverse(F);
  c = invF*(b-f);
  C = invF*(B+Q)*~invF;
}

/*! given x~{a,A} and y|x~{Fx+f,Q} and given evidence for y,
    this returns a potential U(x) = {c,C} such that x|y~{c,C}*P(x) given a prior P(x) */
void LinGaussPotential(arr& c, arr& C, arr& y, arr& f, arr& F, arr& Q){
  arr invF=inverse(F);
  C = invF*Q*~invF;
  c = invF*(y-f);
}

/*! given x~{a,A} and y|x~{f(x),Q} and given evidence for y,
    this returns a potential U(x) = {c,C} such that x|y~{c,C}*P(x) given a prior P(x) */
void FctGaussPotential(arr& c, arr& C, arr& y, Trans b, arr& Q){
  c=y; C=Q;
  unscentedTransform(c,C,b);
}

//! given x~{a,A} and y|x~{Fx+f,Q}, this returns the joint (x,y)~{c,C}
void jointFwd(arr& c, arr& C, arr& a, arr& A, arr& f, arr& F, arr& Q){
  arr FA=F*A,tFA=~FA;
  blockMatrix(C, A, tFA, FA, Q + F*tFA);
  blockVector(c, a, F*a + f);
}

/*! given a joint (x,y)~{c,C} and a potential {b,B} for y|evidence,
    this returns the updated joint (x,y)|evidence~{d,D} */
void multiplyToJoint(arr& d, arr& D, arr& c, arr& C, arr& b, arr& B){
  uint i,j,n2=b.N,n1=c.N-b.N;
  arr invC,invB,invD,tmp;
  invC=inverse(C);
  invB=inverse(B);
  //construct sum of both
  invD=invC;
  for(i=0;i<n2;i++) for(j=0;j<n2;j++){ //loop over the lower right blockmatrix
    invD(n1+i,n1+j) += invB(i,j);
  }
  D = inverse(invD);
  d.resize(n1);
  d.setZero();
  d.append(invB*b); //now is a block vector d = [0 invB*b]
  d = D * (invC * c + d);
}

void getLinFwdFromJoint(arr& f, arr& F, arr& Q, uint n1, uint n2, arr& j, arr& J){
  CHECK(n1+n2==j.N,"");
  arr A,B,C,invA,a,b;
  a=j.sub(0,n1-1);        //upper vector
  b=j.sub(n1,-1);         //lower vector
  A=J.sub(0,n1-1,0,n1-1); //upper left
  B=J.sub(n1,-1,n1,-1);   //lower right
  C=J.sub(0,n1-1,n1,-1);  //upper right
  invA=inverse(A);

  F = ~C*invA;
  f = b - F*a;
  Q = B - F*C;
}

void estimate(arr& X,arr& mean,arr& cov){
  uint k=X.d0;
  arr ones(k); ones=1.;
  mean=ones*X;
  cov =~X*X;
  cov =(cov  - (mean^mean)/(double)k)/double(k);
  mean/=(double)k;
}

void sampleGauss(arr& x,arr& A){
  x.resize(A.d0);
  rndGauss(x);
  arr U,V;
  svd(A,U,V);
  x=U*x;
}
void sampleGauss(arr& x,arr& a,arr& A){
  CHECK(A.nd==2 && A.d0==A.d1 && a.nd==1 && a.N==A.d0,"");
  sampleGauss(x,A);
  x+=a;
}

void sampleGauss(arr& X,arr& a,arr& A,uint N){
  uint i;
  X.resize(N,a.N);
  for(i=0;i<N;i++) sampleGauss(X[i](),a,A);
}

#endif
