void GaussianProcess::calculate(){
  uint i,j,N=Y.N;
  doubleA gram(N,N),xi,xj;
  if(N){ xi.referToSubDim(X,0); xj.referToSubDim(X,0); }
  for(i=0;i<N;i++){ xi.redirect(X,i); gram(i,i) = kernelF(kernelP,xi,xi); }
  for(i=1;i<N;i++){
    xi.redirect(X,i);
    for(j=0;j<i;j++){
      xj.redirect(X,j);
      gram(i,j) = gram(j,i) = kernelF(kernelP,xi,xj);
    }
  }
  inverse(invGram,gram);
  if(N) invGY = invGram * Y;
}

void GaussianProcess::update(const doubleA& x,double& y){
  uint i,N=X.N;

  static arr k,m,M,Xi;
  //update of inverse Gram matrix:
  if(false & N){
    double mu;
    k.resize(N); m.resize(N); M.resize(N,N); Xi.referToSubDim(X,0);
    for(i=0;i<N;i++){ Xi.redirect(X,i); k(i)=kernelF(kernelP,x,Xi); }
    innerProduct(m,invGram,k);
    mu=1./( kernelF(kernelP,x,x) - scalarProduct(k,m) );
    m *= -mu;
    M = invGram;
    M += (1./mu) * m^m;
    
    blockMatrix(invGram, M, m, m, mu);
    
    X.append(x); //append it to the data
    Y.append(y);
  }else{
    X.reshape(N,x.N);
    Y.reshape(N);
    X.append(x); //append it to the data
    Y.append(y);
    calculate();
  }
  invGY = invGram * Y;
#if MT_GP_DEBUG
  arr iG=invGram;
  calculate();
  double err=maxDiff(iG,invGram);
  CHECK(err<1e-6,"mis-updated inverse Gram matrix" <<err <<endl <<iG <<invGram);
#endif
}

void GaussianProcess::evaluate(const doubleA& x,double& y,double& sig){
  uint i,N=Y.N;
  static arr k,Xi,invGk;
  if(N){
    if(k.N!=N) k.resize(N); if(!Xi.N) Xi.referToSubDim(X,0);
    for(i=0;i<N;i++){ Xi.redirect(X,i); k(i)=kernelF(kernelP,x,Xi); }
    y = scalarProduct(k,invGY);
    innerProduct(invGk,invGram,k);
    sig = kernelF(kernelP,x,x) - scalarProduct(k,invGk);
  }else{
    y = 0.; sig=kernelF(kernelP,x,x);
  }
  if(sig<0) sig=0.; else sig = ::sqrt(sig);
}

void GaussianProcess::gradient(doubleA& grad,const doubleA& x){
  CHECK(X.N && x.N==X.d1,"can't calculate gradient without data");
  uint i,N=Y.N;
  doubleA dk;
  static arr Xi; Xi.referToSubDim(X,0);
  grad.resize(x.N);
  grad.setZero();
  for(i=0;i<N;i++){
    Xi.redirect(X,i);
    dkernelF(dk,kernelP,x,Xi);
    grad += invGY(i) * dk;
  }
}

void GaussianProcess::evaluate(const doubleA& X,doubleA& Y,doubleA& S){
  uint i;
  static arr Xi;
  Y.resize(X.d0); S.resize(X.d0); Xi.referToSubDim(X,0);
  for(i=0;i<X.d0;i++){ Xi.redirect(X,i); evaluate(Xi,Y(i),S(i)); }
}
