/*  Copyright (C) 2000, 2006  Marc Toussaint (mtoussai@inf.ed.ac.uk)
    under the terms of the GNU LGPL (http://www.gnu.org/copyleft/lesser.html)
    see the `std.h' file for a full copyright statement  */

#include "opengl.h"
//#include "target.h"

void glStandardLight(){
  static GLfloat ambient[]   = { 0.0, 0.0, 0.0, 1.0 };
  static GLfloat diffuse[]   = { 1.0, 1.0, 1.0, 1.0 };
  static GLfloat specular[]  = { 1.0, 1.0, 1.0, 1.0 };
  static GLfloat position[]  = { 1000.0, -800.0, 1000.0, 1.0 };
  static GLfloat direction[] = { -1.0, .8, -1.0 };
  glEnable(GL_LIGHTING);
  glEnable(GL_LIGHT0);
  glLightfv(GL_LIGHT0, GL_AMBIENT, ambient);
  glLightfv(GL_LIGHT0, GL_DIFFUSE, diffuse);
  glLightfv(GL_LIGHT0, GL_SPECULAR, specular);
  glLighti (GL_LIGHT0, GL_SPOT_CUTOFF, 90);
  glLighti (GL_LIGHT0, GL_SPOT_EXPONENT, 5);
  glLightfv(GL_LIGHT0, GL_POSITION, position);
  glLightfv(GL_LIGHT0, GL_SPOT_DIRECTION, direction);
}

// Henryk's openGL color tab, 4. Mar 06 (hh)
void glColor(int col){
  static const GLfloat colorsTab[6][4] = {{0.2, 0.2, 1.0, 1.0}, // blue
					{1.0, 0.8, 0.0, 1.0}, // gold
					{1.0, 0.0, 0.0, 1.0}, // red
					{0.7, 0.7, 0.7, 1.0}, // gray
					{1.0, 1.0, 1.0, 1.0}, // white
					{0.2, 1.0, 0.2, 1.0}}; // green

  if (col<0) col=0; if (col>5) col=5;
  glColor(colorsTab[col][0],colorsTab[col][1],colorsTab[col][2],colorsTab[col][3]);
}

void glColor(float r, float g, float b, float alpha){
  float amb=1.f,diff=1.f,spec=1.f;
  GLfloat ambient[4],diffuse[4],specular[4];
  ambient[0] = r*amb;
  ambient[1] = g*amb;
  ambient[2] = b*amb;
  ambient[3] = alpha;
  diffuse[0] = r*diff;
  diffuse[1] = g*diff;
  diffuse[2] = b*diff;
  diffuse[3] = alpha;
  specular[0] = spec*.5*(1.+r);
  specular[1] = spec*.5*(1.+g);
  specular[2] = spec*.5*(1.+b);
  specular[3] = alpha;
#if 0
  glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE, diffuse);
#else
  glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT, ambient);
  glMaterialfv(GL_FRONT_AND_BACK, GL_DIFFUSE, diffuse);
  glMaterialfv(GL_FRONT_AND_BACK, GL_SPECULAR, specular);
  glMaterialf (GL_FRONT_AND_BACK, GL_SHININESS, 50.0f);
#endif
  glColor3f(r,g,b);
}


/* // shadows do not work with a light source;
   // thus, we need to leave this out. 4. Mar 06 (hh)
void glShadowTransform()
{
  GLfloat matrix[16];
  for (int i=0; i<16; i++) matrix[i] = 0;
  matrix[0]=1;
  matrix[5]=1;
  matrix[8]=-1;  //light_x
  matrix[9]=-1;  //light_y
  matrix[14]=.02; //ground offset
  matrix[15]=1;
  glPushMatrix();
  glMultMatrixf (matrix);
}
*/

void glTransform(const double pos[3], const double R[12]){
    GLfloat matrix[16];
    matrix[0]=R[0];
    matrix[1]=R[4];
    matrix[2]=R[8];
    matrix[3]=0;
    matrix[4]=R[1];
    matrix[5]=R[5];
    matrix[6]=R[9];
    matrix[7]=0;
    matrix[8]=R[2];
    matrix[9]=R[6];
    matrix[10]=R[10];
    matrix[11]=0;
    matrix[12]=pos[0];
    matrix[13]=pos[1];
    matrix[14]=pos[2];
    matrix[15]=1;
    glPushMatrix();
    glMultMatrixf(matrix);
}

void glDrawText(const char* txt,float x,float y,float z){
#if defined __glut_h__ || (defined __FREEGLUT_H__ && !defined MT_Cygwin)
  glRasterPos3f(x,y,z);
  void *font=GLUT_BITMAP_HELVETICA_12;
  while(*txt){ 
    switch(*txt){
    case '\n':
      y+=15;
      glRasterPos3f(x,y,z);
      break;
    case '\b':
      if(font==GLUT_BITMAP_HELVETICA_12) font=GLUT_BITMAP_HELVETICA_18;
      else font=GLUT_BITMAP_HELVETICA_12;
      break;
    default:
      glutBitmapCharacter(font,*txt); 
    }
    txt++;
  }
#endif
}

void glDrawRect(float x1,float y1,float z1,float x2,float y2,float z2,
	        float x3,float y3,float z3,float x4,float y4,float z4){
  glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);
  glBegin(GL_POLYGON);
  glVertex3f(x1,y1,z1);
  glVertex3f(x2,y2,z2);
  glVertex3f(x3,y3,z3);
  glVertex3f(x4,y4,z4);
  glVertex3f(x1,y1,z1);
  glEnd();
}

void glDrawRect(float x1,float y1,float z1,float x2,float y2,float z2,
	        float x3,float y3,float z3,float x4,float y4,float z4,
		float r,float g,float b){
  glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);
  glBegin(GL_POLYGON);
  glColor(r,g,b);
  glVertex3f(x1,y1,z1);
  glVertex3f(x2,y2,z2);
  glVertex3f(x3,y3,z3);
  glVertex3f(x4,y4,z4);
  glVertex3f(x1,y1,z1);
  glEnd();
}

void glDrawRect(float x,float y,float z,float r){
  glDrawRect(x-r,y-r,z,x-r,y+r,z,x+r,y+r,z,x+r,y-r,z);
}

void glDrawFloor(float x,float r,float g,float b){
  x/=2.;
  glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);
  glColor(r,g,b);
  glBegin(GL_POLYGON);
  glNormal3f(0,0,1);
  glVertex3f(-x,-x,0.);
  glVertex3f( x,-x,0.);
  glVertex3f( x, x,0.);
  glVertex3f(-x, x,0.);
  glVertex3f(-x,-x,0.);
  glEnd();
	

  float grey=0.75;
  glColor(grey,grey,grey);
  for(float i=0.0; i<= 1.1;i= i+0.1){
    for(float j=0.0; j <= 1.1;j=j+0.1){
      glBegin(GL_LINES);
      glVertex3f(i,-1,0.001);
      glVertex3f(i,1,0.001);
      glEnd();
      glBegin(GL_LINES);
      glVertex3f(-i,-1,0.001);
      glVertex3f(-i,1,0.001);
      glEnd();
      glBegin(GL_LINES);
      glVertex3f(-1,i,0.001);
      glVertex3f(1,i,0.001);
      glEnd();
      glBegin(GL_LINES);
      glVertex3f(-1,-i,0.001);
      glVertex3f(1,-i,0.001);
      glEnd();
    }
  }

  grey=0.25;
  glColor(grey,grey,grey);
  glBegin(GL_LINE_STRIP);
  glVertex3f(-1,-1,0.002);
  glVertex3f(-1,1,0.002);
  glVertex3f(1,1,0.002);
  glVertex3f(1,-1,0.002);
  glVertex3f(-1,-1,0.002);
  glEnd();
  		
	
    


}

// used Henryk's drawBox routine to get the 
// correct shading,  4. Mar 06 (hh)
void glDrawBox(float x,float y,float z)
{
  static GLfloat n[6][3] =
    {
      {-1.0, 0.0, 0.0},
      {0.0, 1.0, 0.0},
      {1.0, 0.0, 0.0},
      {0.0, -1.0, 0.0},
      {0.0, 0.0, 1.0},
      {0.0, 0.0, -1.0}
    };

  static GLint faces[6][4] =
    {
      {0, 1, 2,3},
      {3, 2, 6, 7},
      {7, 6, 5, 4},
      {4, 5, 1, 0},
      {5, 6, 2, 1},
      {7, 4, 0, 3}
    };
  GLfloat v[8][3];
  GLint i;

  v[0][0] = v[1][0] = v[2][0] = v[3][0] = -x / 2;
  v[4][0] = v[5][0] = v[6][0] = v[7][0] =  x / 2;
  v[0][1] = v[1][1] = v[4][1] = v[5][1] =  -y / 2;
  v[2][1] = v[3][1] = v[6][1] = v[7][1] =  y / 2;
  v[0][2] = v[3][2] = v[4][2] = v[7][2] =  -z / 2;
  v[1][2] = v[2][2] = v[5][2] = v[6][2] =  z / 2;

  for (i = 5; i >= 0; i--) {
    glBegin(GL_QUADS);
    glNormal3fv(&n[i][0]);
    glVertex3fv(&v[faces[i][0]][0]);
    glVertex3fv(&v[faces[i][1]][0]);
    glVertex3fv(&v[faces[i][2]][0]);
    glVertex3fv(&v[faces[i][3]][0]);
    glEnd();
  }
}

void glDrawDimond(float x,float y,float z){
  glBegin(GL_TRIANGLE_FAN);
  glVertex3f(.0,.0, z);
  glVertex3f( x,.0,.0);
  glVertex3f(.0, y,.0);
  glVertex3f(-x,.0,.0);
  glVertex3f(.0,-y,.0);
  glVertex3f( x,.0,.0);
  glEnd();
  glBegin(GL_TRIANGLE_FAN);
  glVertex3f(.0,.0,-z);
  glVertex3f( x,.0,.0);
  glVertex3f(.0, y,.0);
  glVertex3f(-x,.0,.0);
  glVertex3f(.0,-y,.0);
  glVertex3f( x,.0,.0);
  glEnd();
}

void drawCoordinateFrame(void* motionObj)
{
	//motion *m = (motion *)motionObj;

	// x-axis = green
	glBegin(GL_LINE_STRIP);
		glColor(4.0,0.0,0.0);
		glVertex3f(0.0,0.0,0.001);
		glVertex3f(1.0,0.0,0.001);
		glVertex3f(0.0,0.0,0.002);
		glVertex3f(1.0,0.0,0.002);
		glVertex3f(0.0,0.0,0.003);
		glVertex3f(1.0,0.0,0.003);
		glVertex3f(0.0,0.0,0.004);
		glVertex3f(1.0,0.0,0.004);
    glEnd();

	// y-axis = green
	glBegin(GL_LINE_STRIP);
		glColor(0.0,4.0,0.0);
		glVertex3f(0.0,0.0,0.001);
		glVertex3f(0.0,1.0,0.001);
		glVertex3f(0.0,0.0,0.002);
		glVertex3f(0.0,1.0,0.002);
		glVertex3f(0.0,0.0,0.003);
		glVertex3f(0.0,1.0,0.003);
		glVertex3f(0.0,0.0,0.004);
		glVertex3f(0.0,1.0,0.004);
    glEnd();

	// z-axis = blue
	glBegin(GL_LINE_STRIP);
		glColor(0.0,0.0,4.0);
		glVertex3f(0.001,0.0,0.0);
		glVertex3f(0.001,0.0,1.0);
		glVertex3f(0.002,0.0,0.0);
		glVertex3f(0.002,0.0,1.0);
		glVertex3f(0.003,0.0,0.0);
		glVertex3f(0.003,0.0,1.0);
		glVertex3f(0.004,0.0,0.0);
		glVertex3f(0.004,0.0,1.0);
    glEnd();

}

#if 0
void drawTarget(void* tar){
	geo3d::Vector color;
	geo3d::Vector position;
	float radius;
	int  stacks, slices;
	
	target *_target = (target*)tar;

	color = _target->getRGB();
	position = _target->getPosition();
	radius = _target->getRadius();
	stacks = _target->getStacks();
	slices = _target->getSlices();

	glPushMatrix();
	glColor(color[0],color[1],color[2]);
	
	GLUquadric *style=gluNewQuadric();
	glTranslatef(position[0],position[1],position[2]);
	gluSphere(style,radius,slices,stacks); // last two value for detail
	gluDeleteQuadric(style);
	glPopMatrix();


	
	glDrawGridBox(0.0,0.0,0.0,position[0],position[1],position[2]);
}
#endif

void glDrawSphere(float radius){
  GLUquadric *style=gluNewQuadric();
  gluSphere(style,radius,10,10); // last two value for detail
  gluDeleteQuadric(style);
}

// add cylinder, 6. Mar 06 (hh)
void glDrawCylinder(float radius, float length,bool closed){
  GLUquadric *style=gluNewQuadric();
  glTranslatef(0,0,-length/2);
  gluCylinder(style,radius,radius,length,20,1); 
  if(closed){
    gluDisk(style,0,radius,20,1); 
    glTranslatef(0,0,length);
    gluDisk(style,0,radius,20,1); 
    glTranslatef(0,0,-length/2);
  }
  gluDeleteQuadric(style);
}

// add capped cylinder, 6. Mar 06 (hh)
void glDrawCappedCylinder(float radius, float length)
{
  GLUquadric *style1=gluNewQuadric();
  GLUquadric *style2=gluNewQuadric();
  GLUquadric *style3=gluNewQuadric();

  glTranslatef(0,0,-length/2);
  gluCylinder(style1,radius,radius,length,20,1);
  // add two spheres to the cylinder:
  glTranslatef(0,0,length);
  gluSphere(style2,radius,10,10);
  glTranslatef(0,0,-length);
  gluSphere(style3,radius,10,10);

  gluDeleteQuadric(style1);
  gluDeleteQuadric(style2);
  gluDeleteQuadric(style3);
}

void glDrawGridBox(float x=10.){
  x/=2.;
  glBegin(GL_LINE_LOOP);
  glVertex3f(-x,-x,-x);
  glVertex3f(-x,-x, x);
  glVertex3f(-x, x, x);
  glVertex3f(-x, x,-x);
  glEnd();
  glBegin(GL_LINE_LOOP);
  glVertex3f( x,-x,-x);
  glVertex3f( x,-x, x);
  glVertex3f( x, x, x);
  glVertex3f( x, x,-x);
  glEnd();
  glBegin(GL_LINES);
  glVertex3f( x, x, x);
  glVertex3f(-x, x, x);
  glVertex3f( x,-x, x);
  glVertex3f(-x,-x, x);
  glVertex3f( x, x,-x);
  glVertex3f(-x, x,-x);
  glVertex3f( x,-x,-x);
  glVertex3f(-x,-x,-x);
  glEnd();
}

void glDrawGridBox(float x1,float y1,float z1,float x2,float y2,float z2){
  //glDrawFloor(x);
  glBegin(GL_LINE_STRIP);
  glVertex3f(x1,y1,z1+0.001);
  glVertex3f(x2,y1,z1+0.001);
  glVertex3f(x2,y2,z1+0.001);
  glVertex3f(x1,y2,z1+0.001);
  glVertex3f(x1,y1,z1+0.001);
  glEnd();
 
  glBegin(GL_LINES);
   glVertex3f( x2, y2,z1 +0.001);
  glVertex3f( x2, y2,z2 +0.001);
  glEnd();
}

void glDrawKhepera(){
  GLUquadric *style=gluNewQuadric();
  glPushMatrix();
  glRotatef(-90,1,0,0);
  glColor3f (.3,.3,.3);
  gluCylinder(style,1.5,1.5,2,20,1);
  glPopMatrix();

  glColor3f (1,0,0);
  glBegin(GL_LINES);
  glVertex3f(0,2,0);
  glVertex3f(0,2,-2.5);
  glEnd();
  gluDeleteQuadric(style);
}

void glMakeSquare(int num){
  glNewList(num,GL_COMPILE);
  glColor3f(1.,0.,0.);
  glBegin( GL_LINE_LOOP );
  glVertex3f( -1 , -1 , 0. );
  glVertex3f( -1 , +1 , 0. );
  glVertex3f( +1 , +1 , 0. );
  glVertex3f( +1 , -1 , 0. );
  glEnd();
  glEndList();
}

void glMakeStdSimplex(int num){
  glNewList(num,GL_COMPILE);
  //glPolygonMode(GL_BACK,GL_FILL);
  glShadeModel(GL_SMOOTH);
  glBegin(GL_TRIANGLE_FAN);
  glColor3f (1.,1.,1.);
  glVertex3f(0.,0.,0.);
  glColor3f (1.,0.,0.);
  glVertex3f(1.,0.,0.);
  glColor3f (0.,1.,0.);
  glVertex3f(0.,1.,0.);
  glColor3f (0.,0.,1.);
  glVertex3f(0.,0.,1.);
  glColor3f (1.,0.,0.);
  glVertex3f(1.,0.,0.);
  glEnd();
  /*
    glColor4f(.5,.5,.5,.9);
    glBegin(GL_POLYGON);
    glVertex3f( 1. , 0. , 0. );
    glVertex3f( 0. , 1. , 0. );
    glVertex3f( 0. , 0. , 1. );
    glEnd();
  */
  glEndList();
}

void glMakeTorus(int num){
  glNewList(num,GL_COMPILE);

  GLint i, j, rings, sides;
  float theta1, phi1, theta2, phi2;
  float v0[03], v1[3], v2[3], v3[3];
  float t0[03], t1[3], t2[3], t3[3];
  float n0[3], n1[3], n2[3], n3[3];
  float innerRadius=0.4;
  float outerRadius=0.8;
  float scalFac;

  rings = 8;
  sides = 10;
  scalFac=1/(outerRadius*2);

  for(i=0;i<rings;i++){
    theta1 = (float)i * 2.0 * MT_PI / rings;
    theta2 = (float)(i + 1) * 2.0 * MT_PI / rings;
    for(j=0;j<sides;j++){
      phi1 = (float)j * 2.0 * MT_PI / sides;
      phi2 = (float)(j + 1) * 2.0 * MT_PI / sides;
	
      v0[0] = cos(theta1) * (outerRadius + innerRadius * cos(phi1));
      v0[1] =-sin(theta1) * (outerRadius + innerRadius * cos(phi1));
      v0[2] = innerRadius * sin(phi1);

      v1[0] = cos(theta2) * (outerRadius + innerRadius * cos(phi1));
      v1[1] =-sin(theta2) * (outerRadius + innerRadius * cos(phi1));
      v1[2] = innerRadius * sin(phi1);

      v2[0] = cos(theta2) * (outerRadius + innerRadius * cos(phi2));
      v2[1] =-sin(theta2) * (outerRadius + innerRadius * cos(phi2));
      v2[2] = innerRadius * sin(phi2);

      v3[0] = cos(theta1) * (outerRadius + innerRadius * cos(phi2));
      v3[1] =-sin(theta1) * (outerRadius + innerRadius * cos(phi2));
      v3[2] = innerRadius * sin(phi2);

      n0[0] = cos(theta1) * (cos(phi1));
      n0[1] =-sin(theta1) * (cos(phi1));
      n0[2] = sin(phi1);

      n1[0] = cos(theta2) * (cos(phi1));
      n1[1] =-sin(theta2) * (cos(phi1));
      n1[2] = sin(phi1);

      n2[0] = cos(theta2) * (cos(phi2));
      n2[1] =-sin(theta2) * (cos(phi2));
      n2[2] = sin(phi2);

      n3[0] = cos(theta1) * (cos(phi2));
      n3[1] =-sin(theta1) * (cos(phi2));
      n3[2] = sin(phi2);

      t0[0] = v0[0]*scalFac + 0.5;
      t0[1] = v0[1]*scalFac + 0.5;
      t0[2] = v0[2]*scalFac + 0.5;

      t1[0] = v1[0]*scalFac + 0.5;
      t1[1] = v1[1]*scalFac + 0.5;
      t1[2] = v1[2]*scalFac + 0.5;

      t2[0] = v2[0]*scalFac + 0.5;
      t2[1] = v2[1]*scalFac + 0.5;
      t2[2] = v2[2]*scalFac + 0.5;

      t3[0] = v3[0]*scalFac + 0.5;
      t3[1] = v3[1]*scalFac + 0.5;
      t3[2] = v3[2]*scalFac + 0.5;

      if((i+j)%2) glColor3f(0.,1.,0.);
      else glColor3f(0.,0.,1.);

      glBegin(GL_POLYGON);
      glNormal3fv(n3); glTexCoord3fv(t3); glVertex3fv(v3);
      glNormal3fv(n2); glTexCoord3fv(t2); glVertex3fv(v2);
      glNormal3fv(n1); glTexCoord3fv(t1); glVertex3fv(v1);
      glNormal3fv(n0); glTexCoord3fv(t0); glVertex3fv(v0);
      glEnd();
    }
  }
  glEndList();
}



void glDrawRobotArm(MT::Array<float>& x){
  CHECK(x.N==6,"wrong DOFs for robot arm draw");

  MT::constrain(x(0),0,90);
  MT::constrain(x(1),-30,90);
  MT::constrain(x(2),-30,90);
  MT::constrain(x(3),0,160);
  MT::constrain(x(4),0,160);
  MT::constrain(x(5),-90,90);

  glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);
  glColor3f(.3,.3,.3);
  glPushMatrix();
  glTranslatef(0,0,10);
  glDrawBox(10,10,20);

  glColor3f(.8,.0,.0);
  glTranslatef(5,0,5);
  glRotatef(x(0),0,1,0);
  glRotatef(x(1),0,0,1);
  glRotatef(x(2),1,0,0);
  glTranslatef(5,0,0);
  glDrawBox(10,5,5);

  glColor3f(.0,.0,.8);
  glTranslatef(5,0,0);
  glRotatef(x(3),0,0,1);
  glTranslatef(5,0,0);
  glDrawBox(10,5,5);

  glColor3f(.8,.8,.0);
  glTranslatef(5,0,0);
  glRotatef(x(4),1,0,0);
  glRotatef(x(5),0,1,0);
  glTranslatef(2,0,0);
  glDrawBox(4,5,2);

  glPopMatrix();
}

#if 0
// dm 08.06.2006
void glDrawTrimesh(geo3d::TriMesh mesh) 
{

		glBegin(GL_TRIANGLES);
	glShadeModel(GL_SMOOTH);							
	for(int i=1;i<mesh.T.d0;i++){
			
		  if(mesh.trinormals) glNormal3dv(mesh.Tn(i));
		  if(!mesh.trinormals) glNormal3dv(mesh.Tn(i));//(mesh.Vn(mesh.T(i).nindices[0]));
		////  if(mesh.colored) glColor3fv(mesh.C(T(i,0)));
		  glVertex3dv(mesh.V(mesh.T(i).vindices[0]));

		  if(mesh.trinormals) glNormal3dv(mesh.Tn(i));
		  if(!mesh.trinormals) glNormal3dv(mesh.Tn(i));//glNormal3dv(mesh.Vn(mesh.T(i).nindices[1]));
		////  if(mesh.colored) glColor3fv(mesh.C(T(i,1)));
		  glVertex3dv(mesh.V(mesh.T(i).vindices[1]));

		  if(mesh.trinormals) glNormal3dv(mesh.Tn(i));
		  if(!mesh.trinormals) glNormal3dv(mesh.Tn(i));//glNormal3dv(mesh.Vn(mesh.T(i).nindices[2]));
		////  if(mesh.colored) glColor3fv(mesh.C(T(i,0)));
		  glVertex3dv(mesh.V(mesh.T(i).vindices[2]));
		  
		
	}
	glEnd();

}
#endif

uint glImageTexture(const byteA &img){
  GLuint texName;

  glPixelStorei(GL_UNPACK_ALIGNMENT, 1);
  glGenTextures(1, &texName);

  glBindTexture(GL_TEXTURE_2D, texName);
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
  switch(img.d2){
  case 0:
  case 1:
    glTexImage2D(GL_TEXTURE_2D, 0, 4, img.d1, img.d0, 0, GL_LUMINANCE, GL_UNSIGNED_BYTE, img.p);
    break;
  case 2:
    glTexImage2D(GL_TEXTURE_2D, 0, 4, img.d1, img.d0, 0, GL_LUMINANCE_ALPHA, GL_UNSIGNED_BYTE, img.p);
    break;
  case 3:
    glTexImage2D(GL_TEXTURE_2D, 0, 4, img.d1, img.d0, 0, GL_RGB, GL_UNSIGNED_BYTE, img.p);
    break;
  case 4:
    glTexImage2D(GL_TEXTURE_2D, 0, 4, img.d1, img.d0, 0, GL_RGBA, GL_UNSIGNED_BYTE, img.p);
    break;
  default:
    HALT("no image fomat");
  }
  return texName;
}

void glDrawTexQuad(uint texture,
		   float x1,float y1,float z1,float x2,float y2,float z2,
	           float x3,float y3,float z3,float x4,float y4,float z4){

  glEnable(GL_TEXTURE_2D);
  glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_DECAL);
  glBindTexture(GL_TEXTURE_2D, texture);
  glBegin(GL_QUADS);
  glTexCoord2f(0.0, 0.0); glVertex3f(x1,y1,z1);
  glTexCoord2f(0.0, 1.0); glVertex3f(x2,y2,z2);
  glTexCoord2f(1.0, 1.0); glVertex3f(x3,y3,z3);
  glTexCoord2f(1.0, 0.0); glVertex3f(x4,y4,z4);
  glEnd();
  glFlush();
  glDisable(GL_TEXTURE_2D);
}

/*! return the RGBA-image of scenery drawn just before; the image
  buffer has to have either 2 dimensions [width,height] for a
  gray-scale luminance image or 3 dimensions [width,height,4] for an
  RGBA-image. */
void glGrabImage(byteA& image){
  CHECK(image.nd==2 ||image.nd==3,"not an image format");
  GLint w=image.d1,h=image.d0;

  //glPixelStorei(GL_PACK_SWAP_BYTES,0); 
  switch(image.d2){
  case 0:
  case 1:
    glPixelTransferf(GL_RED_SCALE,.3333);
    glPixelTransferf(GL_GREEN_SCALE,.3333);
    glPixelTransferf(GL_BLUE_SCALE,.3333);
    glReadPixels(0,0,w,h,GL_LUMINANCE,GL_UNSIGNED_BYTE,image.p);
    glPixelTransferf(GL_RED_SCALE,1.);
    glPixelTransferf(GL_GREEN_SCALE,1.);
    glPixelTransferf(GL_BLUE_SCALE,1.);
    break;
  case 2:
    //glReadPixels(0,0,w,h,GL_GA,GL_UNSIGNED_BYTE,image.p);
    break;
  case 3:
    glReadPixels(0,0,w,h,GL_RGB,GL_UNSIGNED_BYTE,image.p);
    break;
  case 4:
#if defined MT_SunOS
    glReadPixels(0,0,w,h,GL_ABGR_EXT,GL_UNSIGNED_BYTE,image.p);
#else
#if defined MT_Cygwin
    glReadPixels(0,0,w,h,GL_RGBA,GL_UNSIGNED_BYTE,image.p);
#else
    glReadPixels(0,0,w,h,GL_BGRA_EXT,GL_UNSIGNED_BYTE,image.p);
#endif
#endif
    break;
  default: HALT("wrong image format");
  }
}
