/*  Copyright (C) 2000, 2006  Marc Toussaint (mtoussai@inf.ed.ac.uk)
    under the terms of the GNU LGPL (http://www.gnu.org/copyleft/lesser.html)
    see the `std.h' file for a full copyright statement  */

#ifndef MT_opengl_h
#define MT_opengl_h

#include "std.h"
#include "array.h"
#include "geo3d.h"
#include <GL/gl.h>
#include <GL/glu.h>

//---------- utility functions I use as abbreviations
void glStandardLight();
void glColor(float r, float g, float b,float a=1.f);
void glColor(int col);
void glDrawText(const char* txt,float x,float y,float z);
//void glShadowTransform();
void glTransform (const double pos[3], const double R[12]);
void glDrawRect(float x1,float y1,float z1,float x2,float y2,float z2,
		float x3,float y3,float z3,float x4,float y4,float z4,
		float r,float g,float b);
void glDrawRect(float x1,float y1,float z1,float x2,float y2,float z2,
		float x3,float y3,float z3,float x4,float y4,float z4);
void glDrawRect(float x,float y,float z,float rad);
void glDrawFloor(float x,float r,float g,float b);
void glDrawBox(float x,float y,float z);
void glDrawDimond(float x,float y,float z);
void glDrawSphere(float radius);
void glDrawCylinder(float radius, float length,bool closed=true);
void glDrawCappedCylinder(float radius, float length);
void glDrawGridBox(float x);
void glDrawGridBox(float x1,float y1,float z1,float x2,float y2,float z2);
void glDrawKhepera();
void glMakeSquare(int num);
void glMakeStdSimplex(int num);
void glMakeTorus(int num);
void glDrawRobotArm(float a,float b,float c,float d,float e,float f);
uint glImageTexture(const byteA &img);
void glDrawTexQuad(uint texture,
		   float x1,float y1,float z1,float x2,float y2,float z2,
	           float x3,float y3,float z3,float x4,float y4,float z4);
/*! return the RGBA-image of scenery drawn just before; the image
  buffer has to have either 2 dimensions [width,height] for a
  gray-scale luminance image or 3 dimensions [width,height,4] for an
  RGBA-image. */
void glGrabImage(byteA& image);

//--------- 

#ifdef MT_FREEGLUT
#  include "opengl_freeglut.h"
#elif defined MT_QT
#  include "opengl_qt.h"
#else

class OpenGL{
public:
  geo3d::Camera camera;
  MT::String text;

  template<class T>
  void add(const T& obj){}
  void add(void (*staticDraw)(void*),void* classP){ MT_MSG("Warning: adding to dummy OpenGL class"); }
  OpenGL(void* parent=0,const char* name="MT/opengl.h",int width=400,int height=400,int posx=0,int posy=0){}
  //void show();
  bool update(){  MT_MSG("Warning: updating dummy OpenGL class"); return false; };
  int timedupdate(uint millisec){ return 0; };
  bool display(){ MT_MSG("Warning: displaying dummy OpenGL class"); return false; };
  bool watch(){   MT_MSG("Warning: watching dummy OpenGL class"); return false; };
};

#endif

/*
#ifdef MT_MinGW
#include "opengl4.h"
#endif
*/



//--------- implementation

#ifdef MT_IMPLEMENTATION
#include "opengl.cpp"
#ifdef MT_GL2PS
#  include<gl2ps.c>
#endif
#endif

#endif
