/*  Copyright (C) 2000, 2006  Marc Toussaint (mtoussai@inf.ed.ac.uk)
    under the terms of the GNU LGPL (http://www.gnu.org/copyleft/lesser.html)
    see the `std.h' file for a full copyright statement  */

#ifndef MT_opengl_freeglut_h
#define MT_opengl_freeglut_h

#define MT_GL

#include "std.h"
#include "array.h"
#include "geo3d.h"
#include <GL/gl.h>
#include <GL/glu.h>

#define FREEGLUT_STATIC
#include <GL/freeglut.h>
extern "C"{
  void FGAPIENTRY glutMainLoopMT( void );
  void fgDeinitialize( void );
}


class OpenGL;
static MT::Array<OpenGL*> glwins;
extern uint nrWins;


// little struct to store objects to be drawn by one OpenGL instance
struct GLDrawer{ void *classP; void (*staticDraw)(void*); };

struct glEvent{ int button,key,updown,x,y; float dx,dy; void set(int b,int k,int ud,int _x,int _y,float _dx,float _dy){ button=b; key=k; updown=ud; x=_x; y=_y; dx=_dx; dy=_dy; } };

class OpenGL{
public:
  int freeglut_window;

  MT::Array<GLDrawer> drawers; //!< the lists of objects to be drawn
  String text;
  float clearR,clearG,clearB,clearA;  // colors of the beackground (called in glClearColor(...))
  geo3d::Camera camera;
  geo3d::Frame frame;
  int argc;
  char *argv[1];
  bool reportEvents;
  glEvent lastEvent;
  GLuint *selected;
  int pressedkey; //!< stores the key pressed
  double mouseposx,mouseposy; //!< x- and y-position of mouse


  OpenGL(const char* title="MT::OpenGL (freeglut)",int w=400,int h=400,int posx=100,int posy=100){
    argc=1;
    argv[0]="x";

    reportEvents=false;
    selected=0;

    camera.setPosition(0.,0.,10.);
    camera.focus(0,0,0);
    camera.setZRange(1.,1000.);
    camera.setHeightAngle(12.);

    clearR=clearG=clearB=.8;

    if(!nrWins){
      glutInit(&argc, argv);
      glutSetOption(GLUT_ACTION_ON_WINDOW_CLOSE,GLUT_ACTION_CONTINUE_EXECUTION);
    }
    nrWins++;

    glutInitWindowSize(w,h);
    glutInitWindowPosition(posx,posy);
    glutInitDisplayMode(GLUT_RGB | GLUT_DOUBLE | GLUT_DEPTH);

    freeglut_window = glutCreateWindow(title);

    //OpenGL initialization
    //two optional thins:
    glEnable(GL_DEPTH_TEST);
    //glEnable(GL_CULL_FACE); glFrontFace(GL_CW);
    glDepthFunc(GL_LESS);
    glShadeModel(GL_SMOOTH);

    if(glwins.N<(uint)freeglut_window+1) glwins.resizeCopy(freeglut_window+1);
    glwins(freeglut_window) = this;

    glutDisplayFunc( _Draw );
    glutKeyboardFunc( _Key );
    glutMouseFunc ( _Mouse ) ;
    glutMotionFunc ( _Motion ) ;
    glutPassiveMotionFunc ( _PassiveMotion ) ;
    glutCloseFunc ( _Close ) ;
    glutReshapeFunc( _Reshape );
    //glutMouseFunc (MouseButton);
    //glutMotionFunc (MouseMotion);

    //  glutSpecialFunc( Special );
    //  glutVisibilityFunc( Visibility );
    //  glutKeyboardUpFunc( KeyUp );
    //  glutSpecialUpFunc( SpecialUp );
    //  glutJoystickFunc( Joystick, 100 );
    //  glutMouseWheelFunc ( MouseWheel ) ;
    //  glutEntryFunc ( Entry ) ;
  }
  ~OpenGL(){
    nrWins--;
    if(!nrWins){
      fgDeinitialize();
    }
  }

  void add(void (*staticDraw)(void*),const void* classP=0){
    CHECK(staticDraw!=0,"OpenGL: NULL pointer to drawing routine");
    GLDrawer d; d.classP=(void*)classP; d.staticDraw=staticDraw;
    drawers.append(d);
  }
  template<class T> void add(const T& x){ add(x.staticDraw,&x); }
  void clear(){ drawers.resize(0); }

  void show(){ }
  void display(){ update(); } //MT_MSG("NIY"); }
  void about(){ MT_MSG("NIY"); }
  void saveEPS(char*){ MT_MSG("NIY"); }

  /*! basic OpenGL drawing routine -- first arranges the perspective
      with respect to the geo3d::Camera camera - then calls the external
      drawing rountines in the #drawers# list */
  void Draw(){
    selected=0;

    //clear bufferer
    glViewport(0,0,(GLint)width(),(GLint)height());
    glClearColor(clearR,clearG,clearB,clearA);
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

    //set view matrix
    camera.glSetProjectionMatrix(true);

    //set world matrix
    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();

    if(!drawers.N){
      if(!text.N()) text <<"<nothing to draw>";
    }
    uint i;
    for(i=0;i<drawers.N;i++) (*drawers(i).staticDraw)(drawers(i).classP);

    GLint s;
    glGetIntegerv(GL_MODELVIEW_STACK_DEPTH,&s);
    CHECK(s==1,"OpenGL matrix stack has not depth 1 (pushs>pops)");

    glGetIntegerv(GL_RENDER_MODE,&s);
    if(text.N() && s!=GL_SELECT){
      glMatrixMode(GL_MODELVIEW);
      glLoadIdentity();
      glOrtho(0.,(double)width(),(double)height(),.0,-1.,1.);
      glMatrixMode(GL_PROJECTION);
      glLoadIdentity();
      if(clearR+clearG+clearB>1.){
        glColor(0.0,0.0,0.0); // clear text  5. Mar 06 (hh)
      }else{
        glColor(1.0,1.0,1.0); // white colored text,  5. Mar 06 (hh)
      }
      glDrawText(text,10,20,0);
    }
    glutSwapBuffers();
  }

  void unproject(double &x,double &y,double &z){
    double _x,_y,_z;
    GLdouble modelMatrix[16],projMatrix[16];
    GLint viewPort[4];
    camera.glSetProjectionMatrix(true);
    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();
    glGetDoublev(GL_MODELVIEW_MATRIX,modelMatrix);
    glGetDoublev(GL_PROJECTION_MATRIX,projMatrix);
    glGetIntegerv(GL_VIEWPORT,viewPort);
    gluUnProject(x,y,z, modelMatrix, projMatrix, viewPort, &_x, &_y, &_z);
    x=_x; y=_y; z=_z;
  }

  void Key(unsigned char key, int x, int y);
  void Mouse(int button, int updown, int x, int y);
  void Motion(int x, int y);
  void PassiveMotion(int x, int y);
  void Close(){ }
  void Reshape(int w,int h);

  static void _Draw(){ glwins(glutGetWindow())->Draw(); }
  static void _Key(unsigned char key, int x, int y){ glwins(glutGetWindow())->Key(key,x,y); }
  static void _Mouse(int button, int updown, int x, int y){ glwins(glutGetWindow())->Mouse(button,updown,x,y); }
  static void _Motion(int x, int y){ glwins(glutGetWindow())->Motion(x,y); }
  static void _PassiveMotion(int x, int y){ glwins(glutGetWindow())->PassiveMotion(x,y); }
  static void _Close(){ glwins(glutGetWindow())->Close(); }
  static void _Reshape(int w,int h){ glwins(glutGetWindow())->Reshape(w,h); }

  int timedupdate(uint msec){
    int r=update();
    MT::wait(.001*msec);
    return r;
  }
  bool update(){
    glutPostRedisplay () ;
    glutMainLoopEvent();
    return true;
  }
  int watch(){
    pressedkey=0;
    display();
    glutMainLoopMT();
    return pressedkey;
  }

  int width(){ return glutGet(GLUT_WINDOW_WIDTH); }
  int height(){ return glutGet(GLUT_WINDOW_HEIGHT); }

  void setClearColors(float r,float g,float b,float a){ clearR=r; clearG=g; clearB=b; clearA=a; }

};

#ifdef MT_IMPLEMENTATION
#  include "opengl_freeglut.cpp"
#endif

#endif
