#include "plot.h"

MT::PlotData plotData;

static enum{ pgl, pxfig, pgnu} mode=pgnu;
std::ofstream xfs,gps,gpd;
uint block;

#ifdef MT_GL
void MT::PlotData::glDraw(){
  uint a,i,j;
  
  Color c;

  double x=0.,y=0.,z=0.;

  //light?
  if(light) glStandardLight();
  
  if(drawBox) glDrawBox();
  
  //draw arrays
  for(a=0;a<array.N;a++){
    CHECK(array(a).nd<=2,"can't display 3(or higher)-dim arrays");
    if(array(a).nd==1 || (array(a).nd==2 && array(a).d1==1)){ //1D functions
      c.setIndex(a);
      glColor(c.r,c.g,c.b);

      for(i=1;i<array(a).N;i++){
	glBegin(GL_LINES);
	glVertex3f(2.*(i-1)/(array(a).N-1)-1.,array(a).elem(i-1),0);
	glVertex3f(2.*(i  )/(array(a).N-1)-1.,array(a).elem(i  ),0);
	glEnd();
      }
      glBegin(GL_LINE_LOOP);
      glColor3f(0.,0.,0.);
      glVertex3f(-1,-1,0);
      glVertex3f(-1, 1,0);
      glVertex3f( 1, 1,0);
      glVertex3f( 1,-1,0);
      glEnd();
    }
    if(array(a).nd==2 && array(a).d1==2){ //2D path
      c.setIndex(a);
      glColor(c.r,c.g,c.b);
      glBegin(GL_LINE_STRIP);
      for(i=0;i<array(a).d0;i++){
	glVertex3f(2.*i/(array(a).d0-1)-1.,array(a).operator()(i,0),array(a).operator()(i,1));
      }
      glEnd();
    }
    if(array(a).nd==2 && array(a).d1>2){ //2D landscapes
      uint i,j,X=array(a).d1,Y=array(a).d0;
      c.setIndex(a);
      if(!grid){ //as a mesh
	c.whiten(.5);
	CHECK(Y*X==mesh.V.N,"you must recall display(array) when dimensions changed");
	for(j=0;j<Y;j++) for(i=0;i<X;i++){
	  x=2.*(double)i/(X-1.)-1.;
	  y=2.*(double)j/(Y-1.)-1.;
	  z=array(a)(j,i);
	  mesh.V(j*X+i).set(x,y,z);
	  mesh.C(j*X+i)=c;
	  mesh.C(j*X+i).setTemp2(z);
	}
	mesh.calcNormals();
	glColor3f(c.r,c.g,c.b);
	glColor(c.r,c.g,c.b);
        mesh.glDraw();
      }else{ //as a grid
	c.blacken(.5);
	for(j=0;j<Y;j++){ //along the x-axis
	  glBegin(GL_LINE_STRIP);
	  for(i=0;i<X;i++){
	    x=2.*(double)i/(X-1.)-1.;
	    y=2.*(double)j/(Y-1.)-1.;
	    z=array(a)(j,i);
	    //c.setTemp2(z);
	    glColor3f(c.r,c.g,c.b);
	    glColor(c.r,c.g,c.b);
	    glVertex3f(x,y,z);
	  }
	  glEnd();
	}
	for(i=0;i<X;i++){ //along the y-axis
	  glBegin(GL_LINE_STRIP);
	  for(j=0;j<Y;j++){
	    x=2.*(double)i/(X-1.)-1.;
	    y=2.*(double)j/(Y-1.)-1.;
	    z=array(a)(j,i);
	    //c.setTemp2(z);
	    glColor3f(c.r,c.g,c.b);
	    glColor(c.r,c.g,c.b);
	    glVertex3f(x,y,z);
	  }
	  glEnd();
	}
      }
    }
  }
  
  //draw points
  for(i=0;i<points.N;i++){
    c.setIndex(i);
    glColor(c.r,c.g,c.b);
    //glBegin(GL_LINES);
    if(points(i).nd==2){
      for(j=0;j<points(i).d0;j++){
	if(points(i).d1==1){ x=(double)j; y=points(i)(j,0); z=0.; }
	if(points(i).d1==2){ x=points(i)(j,0); y=points(i)(j,1); z=0.; }
	if(points(i).d1>=3){ x=points(i)(j,0); y=points(i)(j,1); z=points(i)(j,2); }
	//z+=.01*i;
	glPushMatrix();
	glTranslatef(x,y,z);
	glDrawDimond(.02,.02,.02);
	glPopMatrix();
	/*glVertex3f(x-.01,y-.01,z);
	glVertex3f(x+.01,y+.01,z);
	glVertex3f(x-.01,y+.01,z);
	glVertex3f(x+.01,y-.01,z);*/
      }
    }else{
      if(points(i).d0==1){ x=points(i)(0); y=0.; z=0.; }
      if(points(i).d0==2){ x=points(i)(0); y=points(i)(1); z=0.; }
      if(points(i).d0>=3){ x=points(i)(0); y=points(i)(1); z=points(i)(2); }
      //z+=.01*i;
      glPushMatrix();
      glTranslatef(x,y,z);
      glDrawDimond(.02,.02,.02);
      glPopMatrix();
    }
    //glEnd();
  }
  
  //draw lines
  for(i=0;i<lines.N;i++){
    if(colors) c.setIndex(i); else c.setIndex(0);
    glColor(c.r,c.g,c.b);
     

    glBegin(GL_LINE_STRIP);
    for(j=0;j<lines(i).d0;j++){
      if(lines(i).d1==1) glVertex3d((double)j,lines(i)(j,0),0.);
      if(lines(i).d1==2) glVertex3d(lines(i)(j,0),lines(i)(j,1),0);
      if(lines(i).d1>=3) glVertex3d(lines(i)(j,0),lines(i)(j,1),lines(i)(j,2));
    }
    glEnd();
  }
    
  //draw planes
  for(i=0;i<planes.N;i+=4){
    c.setIndex(i/4+1);
    glColor(c.r,c.g,c.b);
    glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);
    glBegin(GL_POLYGON);
    glVertex3f(planes(i  )[0],planes(i  )[1],planes(i  )[2]);
    glVertex3f(planes(i+1)[0],planes(i+1)[1],planes(i+1)[2]);
    glVertex3f(planes(i+2)[0],planes(i+2)[1],planes(i+2)[2]);
    glVertex3f(planes(i+3)[0],planes(i+3)[1],planes(i+3)[2]);
    glEnd();
  }
}
void MT::PlotData::glDrawBox(){
  glColor3f(.7,.7,.7);
  glBegin(GL_LINE_LOOP);
  glVertex3f(-1,-1,-1);
  glVertex3f(-1, 1,-1);
  glVertex3f( 1, 1,-1);
  glVertex3f( 1,-1,-1);
  glEnd();
  glBegin(GL_LINE_LOOP);
  glVertex3f(-1,-1, 1);
  glVertex3f(-1, 1, 1);
  glVertex3f( 1, 1, 1);
  glVertex3f( 1,-1, 1);
  glEnd();
  glBegin(GL_LINES);
  glVertex3f(-1,-1,-1);
  glVertex3f(-1,-1, 1);
  glVertex3f( 1,-1,-1);
  glVertex3f( 1,-1, 1);
  glVertex3f(-1,-1,-1);
  glVertex3f(-1,-1, 1);
  glVertex3f( 1, 1,-1);
  glVertex3f( 1, 1, 1);
  glVertex3f(-1, 1,-1);
  glVertex3f(-1, 1, 1);
  glEnd();
}
#endif


void plotGnuplot(){ mode=pgnu; }
#ifdef MT_GL
void plotOpengl(){ mode=pgl; }
void plotOpengl(bool threeD,double xl,double xh,double yl,double yh,double zl,double zh){
  mode=pgl;
  plotData.initGL();
  if(!threeD){
    plotData.opengl->camera.setPosition(.5*(xh+xl),.5*(yh+yl),5.);
    plotData.opengl->camera.focus      (.5*(xh+xl),.5*(yh+yl),.0);
    plotData.opengl->camera.setHeightAbs(1.2*(yh-yl));
    plotData.opengl->camera.setWHRatio((xh-xl)/(yh-yl));
    plotData.opengl->update();
    plotData.opengl->setClearColors(1.,1.,1.,1.);
  }
  plotData.setRange(xl,xh,yl,yh,zl,zh);
}
#endif
void checkInit(){}

void plotFunction(const doubleA& f,double x0,double x1){
  doubleA X;
  uint i,j;
  if(f.nd==2){
    if(x0 && x1){
      X.resize(f.d0,f.d1+1);
      for(i=0;i<f.d0;i++){ X(i,0)=x0+(x1-x0)*i/(f.N-1); for(j=1;j<X.d1;j++) X(i,j)=f(i,j-1); }
    }else{
      X=f;
    }
  }
  if(f.nd==1){
    if(x0 && x1){
      X.resize(f.N,2);
      for(i=0;i<f.d0;i++){ X(i,0)=x0+(x1-x0)*i/(f.N-1); X(i,1)=f(i); }
    }else{
      X=f;
      X.reshape(X.N,1);
    }
  }
  plotData.lines.append(X);
}

void plotFunction(const doubleA& x,const doubleA& f){
  doubleA X(x.d0,x.d1+1);
  uint i,j;
  for(i=0;i<X.d0;i++){
    for(j=0;j<x.d1;j++) X(i,j)=x(i,j);
    X(i,j)=f(i);
  }
  plotData.lines.append(X);
}

void plotSurface(const doubleA& X){
  plotData.array.append(X);
  plotData.mesh.clear();
  plotData.mesh.colored=true;
  plotData.mesh.V.resize(X.N);
  plotData.mesh.C.resize(X.N);
  plotData.mesh.gridToTriangles(X.d1,X.d0);
  //plotData.mesh.gridToStrips(X.d1,X.d0);
}

void plotPoints(const doubleA& X){
  plotData.points.append(X);
}

void plotLine(const doubleA& X){
  plotData.lines.append(X);
}

void plotPoints(const doubleA& X,const doubleA& Y){
  doubleA P;
  uint i,j;
  if(X.nd==2){
    P.resize(X.d0,X.d1+1);
    for(i=0;i<P.d0;i++){
      for(j=0;j<X.d1;j++) P(i,j)=X(i,j);
      P(i,j)=Y(i);
    }
  }else{
    P.resize(X.d0,2);
    for(i=0;i<P.d0;i++){ P(i,0)=X(i); P(i,1)=Y(i); }
  }
  plotData.points.append(P);
}

void writeGnuplotFiles(){
  uint i,j;

  //openfiles
  MT::open(gps,"z.plc");
  MT::open(gpd,"z.pld");
  block=0;
  gps <<"set data style lines\nset size square\n";
  
  if(plotData.lines.N+plotData.points.N) gps <<"\nplot \\\n";

  //pipe data
  bool ior=MT::IOraw;
  MT::IOraw=true;
  //lines
  for(i=0;i<plotData.lines.N;i++){
    for(j=0;j<plotData.lines(i).d0;j++) gpd <<plotData.lines(i)[j] <<std::endl;
    gpd <<std::endl;
    if(block) gps <<",\\\n";
    gps <<"'z.pld' every :::"<<block<<"::"<<block<<" with l notitle";
    block++;
  }
  //points
  for(i=0;i<plotData.points.N;i++){
    for(j=0;j<plotData.points(i).d0;j++) gpd <<plotData.points(i)[j] <<std::endl;
    gpd <<std::endl;
    if(block) gps <<",\\\n";
    gps <<"'z.pld' every :::"<<block<<"::"<<block<<" with p notitle";
    block++;
  }

  if(plotData.array.N) gps <<"\n\npause mouse\nset dgrid3d\n\nsplot \\\n";

  //surfaces
  for(i=0;i<plotData.array.N;i++){
    uint j,k,X=plotData.array(i).d1,Y=plotData.array(i).d0;
    for(j=0;j<Y;j++) for(k=0;k<X;k++){
      gpd <<2.*(double)k/(X-1.)-1. <<' ' <<2.*(double)j/(Y-1.)-1. <<' ' <<plotData.array(i)(j,k) <<std::endl;
    }
    gpd <<std::endl;
    if(i && block) gps <<",\\\n";
    gps <<"'z.pld' every :::"<<block<<"::"<<block<<" with l notitle";
    block++;
  }
  MT::IOraw=ior;

  //close files
  gps<<"\n\npause mouse" <<std::endl;
  gps.close();
  gpd.close();
}
void plot(bool wait){
  switch(mode){
  case pgnu:
    writeGnuplotFiles();
#ifndef MT_MSVC
    system("gnuplot z.plc");
#else
    system("wgnuplot z.plc");
#endif
    break;
#ifdef MT_GL
  case pgl:
    if(wait) plotData.watch();
    else plotData.display();
    break;
#else
  case pgl:
    HALT("can't plot on OpenGL without MT_GL flag");
    break;
#endif
  case pxfig:
    HALT("not implemented yet");
    break;
  }
}

void plotCovariance(const doubleA& mean,const doubleA& cov){
  if(mean.nd==2){
    for(uint k=0;k<mean.d0;k++) plotCovariance(mean[k],cov[k]);
    return;
  }
  uint d=mean.N;
  if(d==1){
    doubleA d(20,2);
    uint i;
    for(i=0;i<d.d0;i++){ //standard Gaussian
      d(i,0)=5. * ((i+.5)/d.d0 - .5);
      d(i,1)=NNsdv(d(i,0),0,1.);
    }
    for(i=0;i<d.d0;i++){ //standard Gaussian
      d(i,0) = ::sqrt(cov(0,0)) * d(i,0) + mean(0);
      d(i,1) *= 1./::sqrt(cov(0,0));
    }
    plotFunction(d);
  }
  if(d==2){
    doubleA d(101,2),Cov,U,V,w;
    double phi;
    uint i;
    if(cov.d0>2){ Cov=cov.sub(0,1,0,1); }else{ Cov.referTo(cov); }
    for(i=0;i<d.d0;i++){ //standard circle
      phi=MT_2PI*((double)i)/(d.d0-1);
      d(i,0)=cos(phi); d(i,1)=sin(phi);
    }
    svd(Cov,U,w,V);
    for(i=0;i<w.N;i++) w(i)=sqrt(w(i)); //trace of eig^2 becomes N!
    for(i=0;i<d.d0;i++){ mult(d[i](),d[i],w); d[i]=V*d[i]; d(i,0)+=mean(0); d(i,1)+=mean(1); }


    plotData.lines.append(d);
  }
  if(d==3){
    doubleA d(101,2),dd(101,3),Cov,U,V,w;
    double phi;
    uint i;
    //x-y
    Cov=cov.sub(0,1,0,1);
    for(i=0;i<d.d0;i++){ //standard circle
      phi=MT_2PI*((double)i)/(d.d0-1);
      d(i,0)=cos(phi); d(i,1)=sin(phi);
    }
    svd(Cov,U,w,V);
    for(i=0;i<w.N;i++) w(i)=sqrt(w(i)); //trace of eig^2 becomes N!
    for(i=0;i<d.d0;i++){ mult(d[i](),d[i],w); d[i]=V*d[i]; d(i,0)+=mean(0); d(i,1)+=mean(1); }
    for(i=0;i<d.d0;i++){ dd(i,0)=d(i,0); dd(i,1)=d(i,1); dd(i,2)=mean(2); } 
    plotData.lines.append(dd);
    //y-z
    Cov=cov.sub(1,2,1,2);
    for(i=0;i<d.d0;i++){ //standard circle
      phi=MT_2PI*((double)i)/(d.d0-1);
      d(i,0)=cos(phi); d(i,1)=sin(phi);
    }
    svd(Cov,U,w,V);
    for(i=0;i<w.N;i++) w(i)=sqrt(w(i)); //trace of eig^2 becomes N!
    for(i=0;i<d.d0;i++){ mult(d[i](),d[i],w); d[i]=V*d[i]; d(i,0)+=mean(1); d(i,1)+=mean(2); }
    for(i=0;i<d.d0;i++){ dd(i,0)=mean(0); dd(i,1)=d(i,0); dd(i,2)=d(i,1); } 
    plotData.lines.append(dd);
    //x-z
    Cov(0,0)=cov(0,0); Cov(1,0)=cov(2,0); Cov(0,1)=cov(0,2); Cov(1,1)=cov(2,2); 
    for(i=0;i<d.d0;i++){ //standard circle
      phi=MT_2PI*((double)i)/(d.d0-1);
      d(i,0)=cos(phi); d(i,1)=sin(phi);
    }
    svd(Cov,U,w,V);
    for(i=0;i<w.N;i++) w(i)=sqrt(w(i)); //trace of eig^2 becomes N!
    for(i=0;i<d.d0;i++){ mult(d[i](),d[i],w); d[i]=V*d[i]; d(i,0)+=mean(0); d(i,1)+=mean(2); }
    for(i=0;i<d.d0;i++){ dd(i,0)=d(i,0); dd(i,1)=mean(1); dd(i,2)=d(i,1); } 
    plotData.lines.append(dd);
  }
}

void plotVectorField(const doubleA& X,const doubleA& dX){
  checkInit();
  uint i;
  MT::IOraw=true;
  if(X.d1==2 && dX.d1==2){
    for(i=0;i<X.d0;i++){ gpd <<X[i] <<dX[i] <<std::endl; }
  }else if(X.d1==2 && dX.d1==1){
    for(i=0;i<X.d0;i++){ gpd <<X[i] <<" 0 " <<dX(i,0) <<std::endl; }
  }else if(X.d1==1){
    for(i=0;i<X.d0;i++){ gpd <<X(i,0) <<' ' <<.1*i <<' ' <<dX(i,0) <<" 0 " <<std::endl; }
  }
  gpd <<std::endl;
  if(block) gps <<",\\\n";
  gps <<"'z.pld' every :::"<<block<<"::"<<block<<" using 1:2:3:4 with vectors head filled notitle";
  block++;
}

void plotClear(){ plotData.clear(); }

void plotGaussians(MT::Array<Gaussian>& G){
  for(uint k=0;k<G.N;k++) plotCovariance(G(k).c,G(k).C);
}
