/*  Copyright (C) 2000, 2006  Marc Toussaint (mtoussai@inf.ed.ac.uk)
    under the terms of the GNU LGPL (http://www.gnu.org/copyleft/lesser.html)
    see the `std.h' file for a full copyright statement  */

#ifndef MT_plot_h
#define MT_plot_h

#include "array.h"
#include "algos.h"
#include "functions.h"
#include "geo3d.h"
#include "gauss.h"
#if defined MT_GL
#  include "opengl.h"
#else
typedef int OpenGL;
typedef int QWidget;
#endif

namespace MT{
class PlotData{
public:
  OpenGL *opengl;
  Array<doubleA> array;        //display arrays
  Array<doubleA> points;       //display points
  Array<doubleA> lines;        //display lines
  Array<geo3d::Vector> planes; //display planes
  geo3d::Trimesh mesh;
  double lo[3],hi[3];
  bool light,grid,colors,drawBox;	       //display options

public:
  PlotData(){
    opengl=0;
    lo[0]=lo[1]=lo[2]= 0.;
    hi[0]=hi[1]=hi[2]= 1.;
    light=false;
    grid=false;
    colors=true;
    drawBox=false;
  }
  ~PlotData(){
    undisplay();
  }

  //! clears all objects from the lists to be displayed
  void clear(){
    array.clear();
    points.clear();
    lines.clear();
    planes.clear();
  }

  void setRange(double xl,double xh,double yl=-1.,double yh=1.,double zl=-1.,double zh=1.){
    lo[0]=xl; hi[0]=xh;
    lo[1]=yl; hi[1]=yh;
    lo[2]=zl; hi[2]=zh;
  }
  void transBackPoint(double &x,double &y){
    x=(hi[0]-lo[0])*(x+1.)/2. + lo[0];
    y=(hi[1]-lo[1])*(y+1.)/2. + lo[1];
  }

  void reset(){ clear(); lo[0]=lo[1]=lo[2]= 0.; hi[0]=hi[1]=hi[2]= 1.; }
  void plot(doubleA& a){
    array.append(a);
    uint i;
    //if a.resizeCopy(a.N);
    if(a.nd==1 || (a.nd==2 && a.d1==1)){
      lo[0]=lo[1]=hi[1]=0.; hi[0]=1.;
      for(i=0;i<a.N;i++){
	if(a.elem(i)<lo[1]) lo[1]=a.elem(i);
	if(a.elem(i)>hi[1]) hi[1]=a.elem(i);
      }
    }else{
      lo[0]=hi[0]=lo[1]=hi[1]=0.;
      for(i=0;i<a.d1;i++){
	if(a(i,0)<lo[0]) lo[0]=a(i,0);
	if(a(i,0)>hi[0]) hi[0]=a(i,0);
	if(a(i,1)<lo[1]) lo[1]=a(i,1);
	if(a(i,1)>hi[1]) hi[1]=a(i,1);
      }
    }
  }

  /*  void add(const doubleAL& pointList){
    doubleAL::node n;
    forall(n,pointList) addPoint(n->data());
  }*/

  //! ...
  void undisplay(){
    if(opengl) delete opengl;
    opengl=0;
  }

#ifdef MT_GL
  void initGL(const char* name=0,uint width=600,uint height=600,int posx=0,int posy=0){
    if(!opengl){
      MT::initQt();
      opengl=new OpenGL(name,width,height,posx,posy);
      opengl->add(staticDraw,this);
      //opengl->camera.setPosition(.0,.0,5.);
      //opengl->camera.setHeightAbs(1.2*(yh-yl));
      //opengl->camera.setWHRatio((xh-xl)/(yh-yl));
      opengl->setClearColors(1.,1.,1.,1.);
    }
  }

  //! ...
  bool display(const char* name=0,uint width=600,uint height=600,int posx=0,int posy=0){
    initGL(name,width,height,posx,posy);
    //opengl->text <<'['<<lo[0]<<':'<<hi[0]<<',' <<lo[1]<<':'<<hi[1]<<',' <<lo[2]<<':'<<hi[2]<<']';
    opengl->show();
    return opengl->update();
  }

  /*! displays and waits until ENTER stroke in the display */
  void watch(){
    display();
    opengl->watch();
  }

  /*! the static draw routine for OpenGL */
  static void staticDraw(void* classP){ ((PlotData*)classP)->glDraw(); }
  //! the OpenGL draw routine
  void glDraw();
  void glDrawBox();
#endif
};
}

extern MT::PlotData plotData;


void plotGnuplot();
#ifdef MT_GL
void plotOpengl();
void plotOpengl(bool threeD,double xl,double xh,double yl=-1.,double yh=1.,double zl=-1.,double zh=1.);
#endif

void plot(bool wait=true);
void plotClear();
void plotFunction(const doubleA& f,double x0=0.,double x1=0.);
void plotFunction(const doubleA& x,const doubleA& f);
void plotSurface(const doubleA& X);
void plotPoints(const doubleA& X);
void plotLine(const doubleA& X);
void plotPoints(const doubleA& X,const doubleA& Y);
void writeGnuplotFiles();
void plotCovariance(const doubleA& mean,const doubleA& cov);
void plotVectorField(const doubleA& X,const doubleA& dX);
void plotGaussians(MT::Array<Gaussian>& G);

#ifdef MT_IMPLEMENTATION
#  include"plot.cpp"
#endif

#endif

